/*
 * pkey.c
 *
 * Copyright (C) AB Strakt 2001, All rights reserved
 *
 * Public/rivate key handling code, mostly thin wrappers around OpenSSL.
 * See the file RATIONALE for a short explanation of why this module was written.
 *
 */
#include <Python.h>
#define crypto_MODULE
#include "crypto.h"

static char *CVSid = "@(#) $Id: pkey.c,v 1.4 2004/01/24 19:14:44 olson Exp $";

/*
 * This is done every time something fails, so turning it into a macro is
 * really nice.
 *
 * Arguments:   None
 * Returns:     Doesn't return
 */
#define FAIL() \
do {                                    \
    exception_from_error_queue();       \
    return NULL;                        \
} while (0)
    

static char crypto_PKey_generate_key_doc[] = "\n\
Generate a key of a given type, with a given number of a bits\n\
\n\
Arguments: self - The PKey object\n\
           args - The Python argument tuple, should be:\n\
             type - The key type (TYPE_RSA or TYPE_DSA)\n\
             bits - The number of bits\n\
Returns:   None\n\
";

static PyObject *
crypto_PKey_generate_key(crypto_PKeyObj *self, PyObject *args)
{
    int type, bits;
    RSA *rsa;
    DSA *dsa;

    if (!PyArg_ParseTuple(args, "ii:generate_key", &type, &bits))
        return NULL;

    switch (type)
    {
        case crypto_TYPE_RSA:
            if ((rsa = RSA_generate_key(bits, 0x10001, NULL, NULL)) == NULL)
                FAIL();
            if (!EVP_PKEY_assign_RSA(self->pkey, rsa))
                FAIL();
            Py_INCREF(Py_None);
            return Py_None;

        case crypto_TYPE_DSA:
            if ((dsa = DSA_generate_parameters(bits, NULL, 0, NULL, NULL, NULL, NULL)) == NULL)
                FAIL();
            if (!DSA_generate_key(dsa))
                FAIL();
            if (!EVP_PKEY_assign_DSA(self->pkey, dsa))
                FAIL();
            Py_INCREF(Py_None);
            return Py_None;
    }

    PyErr_SetString(crypto_Error, "No such key type");
    Py_INCREF(Py_None);
    return Py_None;
}

static char crypto_PKey_bits_doc[] = "\n\
Returns the number of bits of the key\n\
\n\
Arguments: self - The PKey object\n\
           args - The Python argument tuple, should be empty\n\
Returns: The number of bits of the key.\n\
";

static PyObject *
crypto_PKey_bits(crypto_PKeyObj *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, ":bits"))
        return NULL;

    return PyInt_FromLong(EVP_PKEY_bits(self->pkey));
}

static char crypto_PKey_type_doc[] = "\n\
Returns the type of the key\n\
\n\
Arguments: self - The PKey object\n\
           args - The Python argument tuple, should be empty\n\
Returns: The type of the key.\n\
";

static PyObject *
crypto_PKey_type(crypto_PKeyObj *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, ":type"))
        return NULL;

    return PyInt_FromLong(self->pkey->type);
}

static char crypto_PKey_get_modulus_doc[] = "\n\
Return the modulus of the given public key.\n\
\n\
Arguments: self - The PKey object\n\
           args - The Python argument tuple, should be:\n\
Returns:   None\n\
";

static PyObject *
crypto_PKey_get_modulus(crypto_PKeyObj *self, PyObject *args)
{
    RSA *rsa;
    DSA *dsa;
    BIO *bio;
    BUF_MEM *bptr;
    PyObject *ret;

    if (!PyArg_ParseTuple(args, ":get_modulus"))
        return NULL;

    switch (self->pkey->type)
    {
        case crypto_TYPE_RSA:
	    rsa = EVP_PKEY_get1_RSA(self->pkey);

	    bio = BIO_new(BIO_s_mem());
	    BN_print(bio, rsa->n);
	    BIO_get_mem_ptr(bio, &bptr);
	    ret = PyString_FromStringAndSize(bptr->data, bptr->length);
	    BIO_set_close(bio, BIO_CLOSE);
	    BIO_free(bio);

	    return ret;

        case crypto_TYPE_DSA:
	    dsa = EVP_PKEY_get1_DSA(self->pkey);

	    bio = BIO_new(BIO_s_mem());
	    BN_print(bio, dsa->pub_key);
	    BIO_get_mem_ptr(bio, &bptr);
	    ret = PyString_FromStringAndSize(bptr->data, bptr->length);
	    BIO_set_close(bio, BIO_CLOSE);
	    BIO_free(bio);

	    return ret;
    }

    Py_INCREF(Py_None);
    return Py_None;
}

static char crypto_PKey_public_encrypt_doc[] = "\n\
Encrypt a buffer with the given public key.\n\
\n\
Arguments: self - The PKey object\n\
           args - The Python argument tuple, should be:\n\
		buffer - buffer of data to encrypt\n\
Returns:   Buffer of encrypted data\n\
";

static PyObject *
crypto_PKey_public_encrypt(crypto_PKeyObj *self, PyObject *args)
{
    RSA *rsa;
    DSA *dsa;
    PyObject *ret;
    char *dat;
    int len;

    if (!PyArg_ParseTuple(args, "s#:public_encrypt", &dat, &len))
        return NULL;

    switch (self->pkey->type)
    {
        case crypto_TYPE_RSA:
	{
	    int bsize, pbsize;	/* pbsize the size of the blocks minus padding */
	    int nblocks;
	    char *out;
	    char *ibptr, *obptr;
	    int left;
	    int padsize = 11;
	    
	    rsa = EVP_PKEY_get1_RSA(self->pkey);
	    bsize = RSA_size(rsa);
	    pbsize = bsize - padsize;

	    nblocks = len / pbsize;
	    if (len > nblocks * pbsize)
		nblocks++;

	    out = (char *) malloc(nblocks * bsize);
	    
	    for (ibptr = dat, obptr = out, left = len;
		 left > 0;
		 ibptr += pbsize, obptr += bsize, left -= pbsize)
	    {
		int n = left > pbsize ? pbsize : left;
		int rc = RSA_public_encrypt(n, ibptr, obptr, rsa, RSA_PKCS1_PADDING);
		if (rc < 0)
		{
		    free(out);
		    exception_from_error_queue();
		    return NULL;
		}
	    }

	    ret = PyString_FromStringAndSize(out, nblocks * bsize);

	    return ret;
	}

        case crypto_TYPE_DSA:
	    dsa = EVP_PKEY_get1_DSA(self->pkey);

	    return NULL;
    }

    Py_INCREF(Py_None);
    return Py_None;
}


static char crypto_PKey_private_decrypt_doc[] = "\n\
Decrypt a buffer with the given private key.\n\
\n\
Arguments: self - The PKey object\n\
           args - The Python argument tuple, should be:\n\
		buffer - buffer of data to encrypt\n\
Returns:   Buffer of encrypted data\n\
";
static PyObject *
crypto_PKey_private_decrypt(crypto_PKeyObj *self, PyObject *args)
{
    RSA *rsa;
    DSA *dsa;
    PyObject *ret;
    char *dat;
    int len, olen;

    if (!PyArg_ParseTuple(args, "s#:private_decrypt", &dat, &len))
        return NULL;

    switch (self->pkey->type)
    {
        case crypto_TYPE_RSA:
	{
	    int bsize;
	    int nblocks;
	    char *out;
	    char *ibptr, *obptr;
	    int left;
	    
	    rsa = EVP_PKEY_get1_RSA(self->pkey);
	    bsize = RSA_size(rsa);

	    nblocks = len / bsize;
	    if (len != nblocks * bsize)
	    {
		PyErr_SetString(PyExc_ValueError, "Invalid input block");
		return NULL;
	    }

	    out = (char *) malloc(nblocks * bsize);
	    
	    for (olen = 0, ibptr = dat, obptr = out, left = len;
		 left > 0;
		 ibptr += bsize)
	    {
		int n, rc;
		n = left > bsize ? bsize : left;
		rc = RSA_private_decrypt(n, ibptr, obptr, rsa, RSA_PKCS1_PADDING);
		if (rc < 0)
		{
		    long err;
		    free(out);
		    exception_from_error_queue();
		    return NULL;
		}
		obptr += rc;
		left -= bsize;
		olen += rc;
	    }

	    ret = PyString_FromStringAndSize(out, olen);

	    return ret;
	}

        case crypto_TYPE_DSA:
	    dsa = EVP_PKEY_get1_DSA(self->pkey);

	    return NULL;
    }

    Py_INCREF(Py_None);
    return Py_None;
}


/*
 * ADD_METHOD(name) expands to a correct PyMethodDef declaration
 *   {  'name', (PyCFunction)crypto_PKey_name, METH_VARARGS }
 * for convenience
 */
#define ADD_METHOD(name)        \
    { #name, (PyCFunction)crypto_PKey_##name, METH_VARARGS, crypto_PKey_##name##_doc }
static PyMethodDef crypto_PKey_methods[] =
{
    ADD_METHOD(generate_key),
    ADD_METHOD(get_modulus),
    ADD_METHOD(bits),
    ADD_METHOD(type),
    ADD_METHOD(public_encrypt),
    ADD_METHOD(private_decrypt),
    { NULL, NULL }
};
#undef ADD_METHOD


/*
 * Constructor for PKey objects, never called by Python code directly
 *
 * Arguments: pkey    - A "real" EVP_PKEY object
 *            dealloc - Boolean value to specify whether the destructor should
 *                      free the "real" EVP_PKEY object
 * Returns:   The newly created PKey object
 */
crypto_PKeyObj *
crypto_PKey_New(EVP_PKEY *pkey, int dealloc)
{
    crypto_PKeyObj *self;

    self = PyObject_New(crypto_PKeyObj, &crypto_PKey_Type);

    if (self == NULL)
        return NULL;

    self->pkey = pkey;
    self->dealloc = dealloc;

    return self;
}

/*
 * Deallocate the memory used by the PKey object
 *
 * Arguments: self - The PKey object
 * Returns:   None
 */
static void
crypto_PKey_dealloc(crypto_PKeyObj *self)
{
    /* Sometimes we don't have to dealloc the "real" EVP_PKEY pointer ourselves */
    if (self->dealloc)
        EVP_PKEY_free(self->pkey);

    PyObject_Del(self);
}

/*
 * Find attribute
 *
 * Arguments: self - The PKey object
 *            name - The attribute name
 * Returns:   A Python object for the attribute, or NULL if something went
 *            wrong
 */
static PyObject *
crypto_PKey_getattr(crypto_PKeyObj *self, char *name)
{
    return Py_FindMethod(crypto_PKey_methods, (PyObject *)self, name);
}

PyTypeObject crypto_PKey_Type = {
    PyObject_HEAD_INIT(NULL)
    0,
    "PKey",
    sizeof(crypto_PKeyObj),
    0,
    (destructor)crypto_PKey_dealloc,
    NULL, /* print */
    (getattrfunc)crypto_PKey_getattr,
    NULL, /* setattr */
    NULL, /* compare */
    NULL, /* repr */
    NULL, /* as_number */
    NULL, /* as_sequence */
    NULL, /* as_mapping */
    NULL, /* hash */
};


/*
 * Initialize the PKey part of the crypto sub module
 *
 * Arguments: dict - The crypto module dictionary
 * Returns:   None
 */
int
init_crypto_pkey(PyObject *dict)
{
    crypto_PKey_Type.ob_type = &PyType_Type;
    Py_INCREF(&crypto_PKey_Type);
    PyDict_SetItemString(dict, "PKeyType", (PyObject *)&crypto_PKey_Type);
    return 1;
}

